package me.shetj.kt

import android.content.Context
import android.widget.SeekBar
import me.shetj.player.PermissionListener
import me.shetj.player.PlayerListener
import me.shetj.player.RecordListener
import me.shetj.recorder.BuildConfig
import me.shetj.recorder.mixRecorder.MixRecorder
import me.shetj.recorder.simRecorder.BaseRecorder
import me.shetj.recorder.simRecorder.MP3Recorder
import me.shetj.recorder.util.VolumeConfig

/**
 * 通过背景音乐录制的背景音乐
 */
@JvmOverloads
fun simRecorderBuilder(
    audioSource: BaseRecorder.AudioSource = BaseRecorder.AudioSource.MIC,
    isDebug: Boolean = BuildConfig.DEBUG,
    mMaxTime: Int = 1800 * 1000,
    samplingRate: Int = 44100,
    mp3BitRate: Int = 128,//96(高),32（低）
    mp3Quality: Int = 1,
    permissionListener: PermissionListener? = null,
    recordListener: RecordListener? = null,
    wax: Float = 1f

): MP3Recorder {
    return MP3Recorder(audioSource.type)
        .setMaxTime(mMaxTime)
        .setMp3Quality(mp3Quality)
        .setSamplingRate(samplingRate)
        .setMp3BitRate(mp3BitRate)
        .setPermissionListener(permissionListener)
        .setRecordListener(recordListener)
        .setWax(wax).apply {
            setDebug(isDebug)
        }
}

/**
 * 混合音频的录制
 */
@JvmOverloads
fun mixRecorder(
    audioSource: BaseRecorder.AudioSource = BaseRecorder.AudioSource.MIC,
    channel: BaseRecorder.AudioChannel = BaseRecorder.AudioChannel.STEREO,
    isDebug: Boolean = BuildConfig.DEBUG,
    mMaxTime: Int = 1800 * 1000,
    samplingRate: Int = 44100,
    mp3BitRate: Int = 128,//96(高),32（低）
    mp3Quality: Int = 1,
    permissionListener: PermissionListener? = null,
    recordListener: RecordListener? = null,
    wax: Float = 1f

): MixRecorder {
    return MixRecorder(audioSource.type, channel.type)
        .setMaxTime(mMaxTime)
        .setMp3Quality(mp3Quality)
        .setSamplingRate(samplingRate)
        .setMp3BitRate(mp3BitRate)
        .setPermissionListener(permissionListener)
        .setRecordListener(recordListener)
        .setWax(wax).apply {
            setDebug(isDebug)
        }
}


@JvmOverloads
fun simRecorder(
    context: Context,
    simpleName: BaseRecorder.RecorderType = BaseRecorder.RecorderType.MIX,
    audioSource: BaseRecorder.AudioSource = BaseRecorder.AudioSource.MIC,
    isDebug: Boolean = BuildConfig.DEBUG,
    mMaxTime: Int = 1800 * 1000,
    samplingRate: Int = 44100,
    mp3BitRate: Int = 128,//96(高),32（低）
    mp3Quality: Int = 1,
    channel: BaseRecorder.AudioChannel = BaseRecorder.AudioChannel.STEREO,
    permissionListener: PermissionListener? = null,
    recordListener: RecordListener? = null,
    wax: Float = 1f
): BaseRecorder {
   return simRecorderNoContext(
        simpleName,
        audioSource,
        isDebug,
        mMaxTime,
        samplingRate,
        mp3BitRate,
        mp3Quality,
        channel,
        permissionListener,
        recordListener,
        wax
    ).setContextToVolumeConfig(context)
        .setContextToVolumeConfig(context)
}

/**
 * mix 表示混合
 * sim 不支持单双声道录制
 */
@JvmOverloads
fun simRecorderNoContext(
    simpleName: BaseRecorder.RecorderType = BaseRecorder.RecorderType.MIX,
    audioSource: BaseRecorder.AudioSource = BaseRecorder.AudioSource.MIC,
    isDebug: Boolean = BuildConfig.DEBUG,
    mMaxTime: Int = 1800 * 1000,
    samplingRate: Int = 44100,
    mp3BitRate: Int = 128,//96(高),32（低）
    mp3Quality: Int = 1,
    channel: BaseRecorder.AudioChannel = BaseRecorder.AudioChannel.STEREO,
    permissionListener: PermissionListener? = null,
    recordListener: RecordListener? = null,
    wax: Float = 1f
): BaseRecorder {
    return when (simpleName) {
        BaseRecorder.RecorderType.MIX ->
            mixRecorder(
                audioSource = audioSource,
                isDebug = isDebug,
                channel = channel,
                mMaxTime = mMaxTime,
                mp3Quality = mp3Quality,
                samplingRate = samplingRate,
                mp3BitRate = mp3BitRate,
                permissionListener = permissionListener,
                recordListener = recordListener,
                wax = wax
            )
        BaseRecorder.RecorderType.SIM ->
            simRecorderBuilder(
                audioSource = audioSource,
                isDebug = isDebug,
                mMaxTime = mMaxTime,
                mp3Quality = mp3Quality,
                samplingRate = samplingRate,
                mp3BitRate = mp3BitRate,
                permissionListener = permissionListener,
                recordListener = recordListener,
                wax = wax
            )
    }
}

/**
 * 设置录制成功回调，不管是最大还是收到
 */
fun BaseRecorder.onSuccess(onSuccess: (file: String, time: Long) -> Unit = { _: String, _: Long -> }): BaseRecorder {
    return setRecordListener(onSuccess = onSuccess, autoComplete = onSuccess)
}

/**
 * 设置背景音乐播放,暂停和重新开始播放的监听
 */
fun BaseRecorder.onPlayChange(
    onPause: () -> Unit = {},
    onResume: () -> Unit = {}
): BaseRecorder {
    return setPlayListener(onPause = onPause, onResume = onResume)
}

/**
 * 设置背景音乐播放的监听
 */
fun BaseRecorder.setPlayListener(
    onStart: (url: String, duration: Int) -> Unit = { _: String, _: Int -> },
    onPause: () -> Unit = {},
    onResume: () -> Unit = {},
    onStop: () -> Unit = {},
    onCompletion: () -> Unit = {},
    onError: (throwable: Exception) -> Unit = {},
    onProgress: (current: Int, duration: Int) -> Unit = { _: Int, _: Int -> }
): BaseRecorder {
    setBackgroundMusicListener(object : PlayerListener {
        override fun onStart(url: String, duration: Int) {
            onStart(url, duration)
        }

        override fun onPause() {
            onPause()
        }

        override fun onResume() {
            onResume()
        }

        override fun onStop() {
            onStop()
        }

        override fun onCompletion() {
            onCompletion()
        }

        override fun onError(throwable: Exception) {
            onError(throwable)
        }

        override fun onProgress(current: Int, duration: Int) {
            onProgress(current, duration)
        }
    })
    return this
}

/**
 * 设置录制监听
 */
fun BaseRecorder.setRecordListener(
    onStart: () -> Unit = {},
    onResume: () -> Unit = {},
    onReset: () -> Unit = {},
    onRecording: (time: Long, volume: Int) -> Unit = { _: Long, _: Int -> },
    onPause: () -> Unit = {},
    onRemind: (mDuration: Long) -> Unit = {},
    onSuccess: (file: String, time: Long) -> Unit = { _: String, _: Long -> },
    setMaxProgress: (time: Long) -> Unit = {},
    onError: (e: Exception) -> Unit = {},
    autoComplete: (file: String, time: Long) -> Unit = { _: String, _: Long -> },
    needPermission: () -> Unit = {}
): BaseRecorder {

    setRecordListener(object : RecordListener {
        override fun onStart() {
            onStart()
        }

        override fun onResume() {
            onResume()
        }

        override fun onReset() {
            onReset()
        }

        override fun onRecording(time: Long, volume: Int) {
            onRecording(time, volume)
        }

        override fun onPause() {
            onPause()
        }

        override fun onRemind(mDuration: Long) {
            onRemind(mDuration)
        }

        override fun onSuccess(file: String, time: Long) {
            onSuccess(file, time)
        }

        override fun setMaxProgress(time: Long) {
            setMaxProgress(time)
        }

        override fun onError(e: Exception) {
            onError(e)
        }

        override fun autoComplete(file: String, time: Long) {
            autoComplete(file, time)
        }
    })
    setPermissionListener(object : PermissionListener {
        override fun needPermission() {
            needPermission()
        }
    })
    return this
}

inline fun BaseRecorder.setVolumeSeekBar(
    mSeekBar: SeekBar,
    volumeConfig: VolumeConfig,
    crossinline onSeek: (seekBar: SeekBar, progress: Int) -> Unit
) {
    mSeekBar.max = volumeConfig.getMaxVoice()
    mSeekBar.progress = volumeConfig.getCurVolume()
    mSeekBar.setOnSeekBarChangeListener(object : SeekBar.OnSeekBarChangeListener {
        override fun onProgressChanged(seekBar: SeekBar?, progress: Int, fromUser: Boolean) {
            seekBar?.let {
                onSeek(seekBar, progress)
            }
        }

        override fun onStartTrackingTouch(seekBar: SeekBar?) {
        }

        override fun onStopTrackingTouch(seekBar: SeekBar?) {
            seekBar?.let {
                onSeek(seekBar, seekBar.progress)
            }
        }
    })
}